function _round(num, i)
{
  if (i == 0 || i == undefined)
  {
    return Math.round(num);
  }
  else
  {
    var scale = 10 ** i;
    return Math.round(num * scale) / scale;
  }
}

function apply_scale(layer, box)
{
  box.x( box.org_x * layer.w_scale );
  box.y( box.org_y * layer.h_scale );
  if (box.org_w != 'auto')
    box.width( box.org_w * layer.w_scale );
  if (box.org_h != 'auto')
    box.height( box.org_h * layer.h_scale );
  box.fontSize( box.org_fontSize * layer.font_scale);
  box.strokeWidth( box._strokeWidth * layer.font_scale);
  box.scaled = true;
}

function create_konva(container_wrapper, container_hidden_field, container_id, img_url, fonts, trans)
{
  var lineheight_trans = "Line Height",
      stroke_trans = "Stroke",
      fill_trans = "Fill";
  if (trans && trans['lineheight'])
    lineheight_trans = trans['lineheight'];
  if (trans && trans['stroke'])
    stroke_trans = trans['stroke'];
  else
    stroke_trans = [stroke_trans, stroke_trans];
  var fill_trans = "Fill";
  if (trans && trans['fill'])
    fill_trans = trans['fill'];
  else
    fill_trans = [fill_trans, fill_trans];
  var toolbar_wrapper = $("<div class=\"toolbar-wrapper\"></div>");
  var container_inner_wrapper = $("<div id=\"container-inner-wrapper\"></div>")
  var color_picker = $("<input title=\"" + fill_trans[1] + "\" disabled class=\"set_color color-picker font-color\" value=\"#333333\" type=\"color\"/>");
  var font_picker = $("<select class=\"font-picker\" disabled></select>");
  var fontsize_picker = $("<input type=\"number\" class=\"fontsize-picker\" value=\"20\" disabled />");
  var lineheight_picker  = $("<span class=\"line-height-picker-wrapper\"><img src=\"/assets/line-height-line.svg\" title=\"" + lineheight_trans + "\" alt=\"" + lineheight_trans + "\"><input type=\"number\" step=\"0.1\" class=\"line-height-picker\" value=\"1.0\" disabled /></span>")
  var stroke_checker = $("<span title=\"" + stroke_trans[1] +"\"><input type=\"checkbox\" class=\"stroke-checker\" value=\"1\" disabled />" + stroke_trans[0] + "</span>");
  var stroke_picker = $("<input disabled class=\"set_color color-picker stroke-color hide\" value=\"#333333\" type=\"color\" />");
  var strokeWidth_picker = $("<input type=\"number\" step=\"0.1\" class=\"strokeWidth-picker hide\" value=\"2\" />");
  var fill_checker = $("<span class=\"fill-checker-wrapper hide\" title=\"" + fill_trans[1] +"\"><input type=\"checkbox\" class=\"fill-checker\" value=\"1\" checked />" + fill_trans[0] + "</span>");
  var add_btn = $("<button type=\"button\" class=\"fa fa-add btn btn-info pull-right\" disabled></button>");
  var remove_btn = $("<button type=\"button\" class=\"fa fa-remove btn btn-danger pull-right\" disabled></button>");
  var container_img = $("<img id=\"container-img\" src=\"" + img_url + "\" />");
  var container = $("<div id=\"" + container_id + "\"></div>");
  var layer = new Konva.Layer({draggable: false});
  var hidden_field_name = container_hidden_field.data('name');
  toolbar_wrapper.append(color_picker);
  toolbar_wrapper.append(font_picker);
  toolbar_wrapper.append(fontsize_picker);
  toolbar_wrapper.append(lineheight_picker);
  toolbar_wrapper.append(stroke_checker);
  toolbar_wrapper.append(stroke_picker);
  toolbar_wrapper.append(strokeWidth_picker);
  toolbar_wrapper.append(fill_checker);
  toolbar_wrapper.append(add_btn);
  toolbar_wrapper.append(remove_btn);
  toolbar_wrapper.appendTo(container_wrapper);
  container_inner_wrapper.appendTo(container_wrapper).append(container_img).append(container);

  lineheight_picker = lineheight_picker.find('>input');
  stroke_checker = stroke_checker.find('>input');
  fill_checker = fill_checker.find('>input');

  $.each(fonts, function(i, font){
    font_picker.append(new Option(font, i.toString()));
  });

  font_picker.select2({
     // templateResult: function(el){
     //  var $element = $(el.element),
     //    image = $element.data("image");
     //  return $("<img class='thumbnail' src='" + image + "'/><span class='thumbnail-text'>" + el.text + "</span>");
     // },
     // minimumResultsForSearch: -1,
     // width : 250
  });

  layer.font_scale = null;
  layer.w_scale    = null;
  layer.h_scale    = null;

  container_img.on("load", function(){
    var img_width = container_img.outerWidth(true),
        img_height = container_img.outerHeight(true);

    var stage = new Konva.Stage({
      container: container_id,
      width: img_width,
      height: img_height
    });
    // add canvas element
    layer.font_scale = img_width / 500;
    layer.img_width = img_width;
    layer.img_height = img_height;
    layer.w_scale = img_width / 100;
    layer.h_scale = img_height / 100;


    layer.text_boxes.forEach(function(box) {
      if (box.scaled)
        return;
      else
      {
        apply_scale(layer, box);
      }
    });
    layer.pending_boxes.forEach(function(args) {
      args[0] = layer;
      create_text_box.apply(null, args);
    });

    layer.pending_boxes = [];

    add_btn.removeAttr('disabled');
    remove_btn.removeAttr('disabled');
    
    stage.add(layer);
  });

  color_picker.on('input', function(layer){
    var target_node = layer.target_node;
    if (target_node)
    {
      var _fill = this.val();
      target_node._fill = _fill;
      if (target_node.is_fill)
        target_node.fill(_fill);
      target_node.hidden_field_dict.fontColor.val(_fill);
      layer.container_hidden_field.trigger("change");
    }
  }.bind(color_picker, layer));

  stroke_picker.on('input', function(layer){
    var target_node = layer.target_node;
    if (target_node)
    {
      var _stroke = this.val();
      target_node._stroke = _stroke;
      if (target_node.is_stroke)
        target_node.stroke(_stroke);
      target_node.hidden_field_dict.stroke.val(_stroke);
      layer.container_hidden_field.trigger("change");
    }
  }.bind(stroke_picker, layer));

  layer.container = container;
  layer.container_hidden_field = container_hidden_field;
  layer.hidden_field_name = hidden_field_name;
  layer.box_idx = -1;
  layer.fonts = fonts;

  layer.color_picker = color_picker;
  layer.font_picker = font_picker;
  layer.fontsize_picker = fontsize_picker;
  layer.lineheight_picker = lineheight_picker;
  layer.stroke_checker = stroke_checker;
  layer.stroke_picker = stroke_picker;
  layer.strokeWidth_picker = strokeWidth_picker;
  layer.fill_checker = fill_checker;
  layer.max_y = null;
  layer.default_fontsize = 20;
  layer.text_boxes = [];
  layer.pending_boxes = [];

  font_picker.on('change', function(fonts, layer){
    var target_node = layer.target_node;
    if (target_node)
    {
      var fontFamily = fonts[Number(this.val())];
      target_node.fontFamily(fontFamily);
      target_node.hidden_field_dict.fontFamily.val(fontFamily);
      layer.container_hidden_field.trigger("change");
    }
  }.bind(font_picker, fonts, layer));

  fontsize_picker.on('change', function(layer){
    var target_node = layer.target_node;
    if (target_node)
    {
      var org_fontSize = Number(this.val());
      var new_fontSize = layer.font_scale * org_fontSize;
      var org_height = target_node.height();
      var newlines = target_node.getText().match(/\n/g);
      var height_diff = new_fontSize * ((newlines ? newlines.length : 0) + 1) - org_height - 2 * target_node.padding();
      target_node.fontSize(new_fontSize);
      target_node.org_fontSize = org_fontSize;
      target_node.hidden_field_dict.fontSize.val(org_fontSize.toString());
      layer.container_hidden_field.trigger("change");
      if (height_diff > 0)
      {
        target_node.height(org_height + height_diff);
      }
    }
  }.bind(fontsize_picker, layer));

  lineheight_picker.on('change', function(layer){
    var target_node = layer.target_node;
    if (target_node)
    {
      var lineHeight = Number(this.val());
      var fontSize = target_node.fontSize();
      var org_height = target_node.height();
      var newlines = target_node.getText().match(/\n/g);
      var height_diff = fontSize * ((newlines ? newlines.length : 0) + 1) * lineHeight  - org_height - 2 * target_node.padding();
      target_node.setLineHeight(lineHeight);
      target_node.hidden_field_dict.l_h.val(lineHeight.toString());
      layer.container_hidden_field.trigger("change");
      if (height_diff > 0)
      {
        target_node.height(org_height + height_diff);
      }
    }
  }.bind(lineheight_picker, layer));

  strokeWidth_picker.on('input', function(layer){
    var target_node = layer.target_node;
    if (target_node)
    {
      var org_strokeWidth = Number(this.val());
      if (isNaN(org_strokeWidth))
        return;
      var new_strokeWidth = layer.font_scale * org_strokeWidth;
      target_node.strokeWidth(new_strokeWidth);
      target_node._strokeWidth = org_strokeWidth;
      target_node.hidden_field_dict.strokeWidth.val(org_strokeWidth.toString());
      layer.container_hidden_field.trigger("change");
    }
  }.bind(strokeWidth_picker, layer));

  stroke_checker.on('change', function(layer){
    var target_node = layer.target_node;
    if (target_node)
    {
      var is_stroke = this.prop("checked");
      if (is_stroke)
      {
        layer.stroke_checker.attr("checked", "");
        layer.stroke_picker.removeClass("hide");
        layer.strokeWidth_picker.removeClass("hide");
        layer.fill_checker.parent().removeClass("hide");
        target_node.stroke(target_node._stroke);
      }
      else
      {
        layer.stroke_checker.removeAttr("checked");
        layer.stroke_picker.addClass("hide");
        layer.strokeWidth_picker.addClass("hide");
        layer.fill_checker.parent().addClass("hide");
        target_node.stroke(null);
      }
      target_node.is_stroke = is_stroke;
      target_node.hidden_field_dict.is_stroke.val(is_stroke.toString());
      layer.container_hidden_field.trigger("change");
    }
  }.bind(stroke_checker, layer));

  fill_checker.on('change', function(layer){
    var target_node = layer.target_node;
    if (target_node)
    {
      var is_fill = this.prop("checked");
      if (is_fill)
      {
        target_node.fill(target_node._fill);
      }
      else
      {
        target_node.fill(null);
      }
      target_node.is_fill = is_fill;
      target_node.hidden_field_dict.is_fill.val(is_fill.toString());
      layer.container_hidden_field.trigger("change");
    }
  }.bind(fill_checker, layer));

  add_btn.on('click', function(layer) {
    create_text_box(layer, {x: 50, y: (layer.max_y || 50) + layer.default_fontsize * 100 / layer.img_width, w: 'auto', h: 'auto'}, "Some text here", layer.default_fontsize);
    layer.container_hidden_field.trigger("change");
  }.bind(add_btn, layer));

  remove_btn.on('click', function(layer) {
    var target_node = layer.target_node;
    if (target_node)
    {
      destroy_text(target_node);
    }
  }.bind(remove_btn, layer));

  container.prop('tabIndex', 1);
  container.focus();

  container.on('keydown', function(layer, event) {
    const key = event.key;
    if (key === "Backspace" || key === "Delete") {
      var target_node = layer.target_node;
      if (target_node)
      {
        destroy_text(target_node);
      }
    }
  }.bind(container, layer));

  layer.resize_func = function(container_img) {
    var img_width, img_height;
    var stage = this.getStage();
    if (this.resized)
    {
      window.setTimeout(function(){
        this.resize_func(container_img);
      }.bind(this), 100);
      return;
    }
    this.resized = true;
    stage.setWidth();
    stage.setHeight();
    img_width = container_img.outerWidth(true);
    img_height = container_img.outerHeight(true);
    stage.setWidth(img_width);
    stage.setHeight(img_height);
    this.font_scale = img_width / 500;
    this.w_scale = img_width / 100;
    this.h_scale = img_height / 100;
    this.img_width = img_width;
    this.img_height = img_height;

    this.pending_boxes = [null];

    this.text_boxes.forEach(function(box) {
      apply_scale(this, box);
    }.bind(this));

    this.pending_boxes = [];

    this.resized = false;
  }.bind(layer, container_img);

  $(window).on("resize", layer.resize_func);

  return layer;
}

function text_box_to_hidden_field(hidden_field_name, idx)
{
  var x = this.org_x.toString();
  var y = this.org_y.toString();
  var w = this.org_w.toString();
  var h = this.org_h.toString();
  var l_h = this.lineHeight().toString();
  var text = this.text();
  var fontSize = this.org_fontSize.toString();
  var fontFamily = this.fontFamily();
  var fontColor = this.fill();
  var is_fill = this.is_fill ? 'true' : 'false';
  var is_stroke = this.is_stroke ? 'true' : 'false';
  var stroke = this._stroke ? this._stroke : '';
  var strokeWidth = this._strokeWidth ? this._strokeWidth : '';
  var hidden_field_dict = {
    "x": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][x]\" value=\"" + x + "\" />"),
    "y": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][y]\" value=\"" + y + "\" />"),
    "w": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][w]\" value=\"" + w + "\" />"),
    "h": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][h]\" value=\"" + h + "\" />"),
    "l_h": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][l_h]\" value=\"" + l_h + "\" />"),
    "text": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][text]\" value=\"" + text + "\" />"),
    "fontSize": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][fontSize]\" value=\"" + fontSize + "\" />"),
    "fontFamily": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][fontFamily]\" value=\"" + fontFamily + "\" />"),
    "fontColor": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][fontColor]\" value=\"" + fontColor + "\" />"),
    "is_fill": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][is_fill]\" value=\"" + is_fill + "\" />"),
    "is_stroke": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][is_stroke]\" value=\"" + is_stroke + "\" />"),
    "stroke": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][stroke]\" value=\"" + stroke + "\" />"),
    "strokeWidth": $("<input type=\"hidden\" name=\"" + hidden_field_name + "[" + idx.toString() + "][strokeWidth]\" value=\"" + strokeWidth + "\" />"),
  };

  Object.keys(hidden_field_dict).forEach(function(dict, k)
  {
    dict[k].appendTo(this.hidden_field_block);
  }.bind(this, hidden_field_dict));

  return hidden_field_dict;
}

function create_text_box(layer, pos, text, fontSize, fontFamily, fontColor, is_stroke, stroke, strokeWidth, is_fill)
{
  if (!layer)
    return;

  if (!layer.font_scale || !layer.w_scale || !layer.h_scale)
  {
    layer.pending_boxes.push(arguments);
    return;
  }
  else if (layer.pending_boxes.length != 0)
  {
    var args = arguments;
    window.setTimeout(function(args){
      create_text_box.apply(this, args);
    }.bind(this, args), 100);
    return;
  }

  var scaled = false;
  var org_fontSize = fontSize,
      org_x = pos.x,
      org_y = pos.y,
      x, y, w, h,
      line_height,
      org_w, org_h,
      _strokeWidth;
  var hidden_field_block = $("<div class=\"hidden-field-block\"></div>");
  var hidden_field_dict;
  if (!fontFamily)
    fontFamily = 'Calibri';

  if (!fontColor)
    fontColor = '#333333';

  if (!stroke)
    stroke = '#333333';

  if (!strokeWidth)
    strokeWidth = 2;

  _strokeWidth = strokeWidth;

  org_w = (pos.w == 0 ? 'auto' : pos.w);
  org_h = (pos.h == 0 ? 'auto' : pos.h);
  x = pos.x;
  y = pos.y;
  w = org_w;
  h = org_h;

  if (layer.font_scale && layer.w_scale && layer.h_scale)
  {
    x = pos.x * layer.w_scale;
    y = pos.y * layer.h_scale;
    if (w != 'auto')
      w *= layer.w_scale;
    if (h != 'auto')
      h *= layer.h_scale;
    fontSize *= layer.font_scale;
    strokeWidth *= layer.font_scale;
    scaled = true;
  }

  if (!is_stroke)
    is_fill = true;

  // create shape
  var textNode = new Konva.Text({
    text: text,
    width: w,
    height: h,
    x: x,
    y: y,
    lineHeight: pos.l_h,
    fontSize: fontSize,
    draggable: true,
    fontFamily: fontFamily,
    stroke: (is_stroke ? stroke : null),
    strokeWidth: (is_stroke ? strokeWidth : null),
    fillAfterStrokeEnabled: true
  });
  textNode.fill((is_fill ? fontColor : false));
  layer.add(textNode);

  textNode._fill = fontColor;
  textNode.scaled = scaled;
  textNode.org_fontSize = org_fontSize;
  textNode.org_x = org_x;
  textNode.org_y = org_y;
  textNode.org_w = org_w;
  textNode.org_h = org_h;
  textNode.is_stroke = is_stroke;
  textNode.is_fill = is_fill;
  textNode._stroke = stroke;
  textNode._strokeWidth = _strokeWidth;
  textNode.hidden_field_block = hidden_field_block;

  if (layer.max_y)
    layer.max_y = Math.max(layer.max_y, pos.y);
  else
    layer.max_y = pos.y;

  layer.box_idx += 1;
  hidden_field_dict = text_box_to_hidden_field.call(textNode, layer.hidden_field_name, layer.box_idx);
  layer.container_hidden_field.append(hidden_field_block);

  textNode.hidden_field_dict = hidden_field_dict;

  // textNode.on('transformend', function () {
  //   console.log('transform end - before reset font size = ' + this.fontSize() + ' at scale ' + this.scaleX());
  //   this.fontSize(this.fontSize() * this.scaleX());
  //   this.scale({x: 1, y: 1});
  //   layer.batchDraw();
  //   console.log('transform end - after reset font size = ' + this.fontSize() + ' at scale ' + this.scaleX());
  // });
  textNode.on('transformend', function (layer) {
    this.org_w = _round(this.width() / layer.w_scale, 5);
    this.org_h = _round(this.height() / layer.h_scale, 5);
    this.hidden_field_dict.w.val(this.org_w.toString());
    this.hidden_field_dict.h.val(this.org_h.toString());
    layer.container_hidden_field.trigger("change");
  }.bind(textNode, layer));

  textNode.on('dragend', function () {
    this.org_x = _round(this.x() / layer.w_scale, 5);
    this.org_y = _round(this.y() / layer.h_scale, 5);
    this.hidden_field_dict.x.val(this.org_x.toString());
    this.hidden_field_dict.y.val(this.org_y.toString());
    layer.container_hidden_field.trigger("change");
  }.bind(textNode));

  textNode.on('transform', function() {
    this.setAttrs({
      width: Math.max(this.width() * this.scaleX(), 5),
      height: Math.max(this.height() * this.scaleY(), 5),
      scaleX: 1,
      scaleY: 1,
    });
  }.bind(textNode));

  var tr_border = new Konva.Transformer({
    // node: textNode,
    keepRatio: false,
    borderEnabled: true,
    borderStrokeWidth: 3,
    rotateEnabled: false,
    enabledAnchors: []
  });

  var tr = new Konva.Transformer({
    // node: textNode,
    ignoreStroke: true,
    rotateEnabled: false,
    // flipEnabled: false,
    keepRatio: false,
    borderEnabled: true,
    // enabledAnchors: ['top-left', 'top-right', 'bottom-left', 'bottom-right'],
    // set minimum width of text
    boundBoxFunc: function (oldBox, newBox) {
      newBox.width = Math.max(30, newBox.width);
      return newBox;
    },
  });

  textNode.children = [tr, tr_border];

  // window.tr_border = tr_border;
  // window.tr = tr;
  // window.layer = layer;

  layer.add(tr_border);
  layer.add(tr);

  // add cursor styling
  textNode.on('mouseover', function(tr_border) {
    document.body.style.cursor = 'pointer';
    if (tr_border.nodes().length == 0)
    {
      tr_border.nodes([this]);
      tr_border.forceUpdate();
    }
  }.bind(textNode, tr_border));
  textNode.on('mouseout', function(tr_border) {
    document.body.style.cursor = 'default';
    if (tr_border.nodes().length != 0)
    {
      tr_border.nodes([]);
      tr_border.forceUpdate();
    }
  }.bind(textNode, tr_border));
  textNode.on('click', function(tr, layer, _, e) {
    if (tr.nodes().length == 0)
    {
      tr.nodes([this]);
      tr.forceUpdate();
      layer.target_node = this;
      layer.color_picker.removeAttr('disabled');
      layer.font_picker.removeAttr('disabled');
      layer.fontsize_picker.removeAttr('disabled');
      layer.lineheight_picker.removeAttr('disabled');
      layer.stroke_checker.removeAttr('disabled');
      layer.stroke_picker.removeAttr('disabled');
      layer.color_picker.val(this._fill);
      layer.font_picker.val(layer.fonts.indexOf(this.fontFamily()).toString());
      layer.fontsize_picker.val(this.org_fontSize.toString());
      layer.lineheight_picker.val(this.lineHeight().toString());
      layer.stroke_checker.prop("checked", this.is_stroke);
      layer.fill_checker.prop("checked", this.is_fill);
      layer.stroke_picker.val(this._stroke);
      layer.strokeWidth_picker.val(this._strokeWidth);
      if (this.is_stroke)
      {
        layer.stroke_checker.attr("checked", "");
        layer.stroke_picker.removeClass("hide");
        layer.strokeWidth_picker.removeClass("hide");
        layer.fill_checker.parent().removeClass("hide");
      }
      else
      {
        layer.stroke_checker.removeAttr("checked");
        layer.stroke_picker.addClass("hide");
        layer.strokeWidth_picker.addClass("hide");
        layer.fill_checker.parent().addClass("hide");
      }
      layer.font_picker.trigger("change");
    }
    this.click_triggered = true;
    layer.click_triggered = true;
  }.bind(textNode, tr, layer));

  textNode.global_click = function(tr, layer, e) {
    if (this.click_triggered) {
      this.click_triggered = false;
      return;
    }
    if (tr.nodes().length != 0)
    {
      tr.nodes([]);
      tr.forceUpdate();
      if (!layer.click_triggered)
      {
        layer.target_node = null;
        layer.color_picker.attr('disabled', '');
        layer.font_picker.attr('disabled', '');
        layer.fontsize_picker.attr('disabled', '');
        layer.lineheight_picker.attr('disabled', '');
        layer.stroke_checker.prop('checked', false);
        layer.stroke_checker.attr('disabled', '');
        layer.stroke_picker.attr('disabled', '');
        layer.stroke_picker.addClass("hide");
        layer.strokeWidth_picker.addClass("hide");
        layer.fill_checker.parent().addClass("hide");
      }
    }
    if (layer.click_triggered)
    {
      layer.click_triggered = false;
    }
  }.bind(textNode, tr, layer);

  layer.container.on('click', textNode.global_click);

  textNode.on('dblclick dbltap', function(tr, layer) {
    var textNode = this;

    // hide text node and transformer:
    this.hide();
    // tr.hide();

    // create textarea over canvas with absolute position
    // first we need to find position for textarea
    // how to find it?

    // at first lets find position of text node relative to the stage:
    var textPosition = this.absolutePosition();

    // so position of textarea will be the sum of positions above:
    var areaPosition = {
      x: layer.container.prop('offsetLeft') + textPosition.x,
      y: layer.container.prop('offsetTop') + textPosition.y,
    };

    // create textarea and style it
    var textarea = document.createElement('textarea');
    layer.container.parent().append(textarea);

    // apply many styles to match text on canvas as close as possible
    // remember that text rendering on canvas and on the textarea can be different
    // and sometimes it is hard to make it 100% the same. But we will try...
    textarea.value = this.text();
    textarea.style.top = areaPosition.y + 'px';
    textarea.style.left = areaPosition.x + 'px';
    textarea.style.width = this.width() - this.padding() * 2 + 'px';
    textarea.style.height =
      this.height() - this.padding() * 2 + 'px';
    textarea.style.fontSize = this.fontSize() + 'px';
    textarea.style.lineHeight = this.lineHeight();
    textarea.style.fontFamily = this.fontFamily();
    textarea.style.textAlign = this.align();
    textarea.style.color = this._fill;
    rotation = this.rotation();
    var transform = '';
    if (rotation) {
      transform += 'rotateZ(' + rotation + 'deg)';
    }

    var px = 0;
    // also we need to slightly move textarea on firefox
    // because it jumps a bit
    var isFirefox =
      navigator.userAgent.toLowerCase().indexOf('firefox') > -1;
    if (isFirefox) {
      px += 2 + Math.round(this.fontSize() / 20);
    }
    transform += 'translateY(-' + px + 'px)';

    textarea.style.transform = transform;

    // reset height
    // textarea.style.height = 'auto';
    // // after browsers resized it we can set actual value
    // textarea.style.height = textarea.scrollHeight + 3 + 'px';

    textarea.focus();

    function removeTextarea() {
      textNode.hidden_field_dict.text.val(textNode.text());
      layer.container_hidden_field.trigger("change");
      textarea.parentNode.removeChild(textarea);
      window.removeEventListener('click', handleOutsideClick);
      // textNode.width('auto');
      textNode.show();
      // tr.show();
      tr.forceUpdate();
    }

    function setTextareaWidth(newWidth) {
      if (!newWidth) {
        // set width for placeholder
        newWidth = this.placeholder.length * this.fontSize();
      }
      // some extra fixes on different browsers
      var isSafari = /^((?!chrome|android).)*safari/i.test(
        navigator.userAgent
      );
      var isFirefox =
        navigator.userAgent.toLowerCase().indexOf('firefox') > -1;
      if (isSafari || isFirefox) {
        newWidth = Math.ceil(newWidth);
      }

      var isEdge =
        document.documentMode || /Edge/.test(navigator.userAgent);
      if (isEdge) {
        newWidth += 1;
      }
      textarea.style.width = newWidth + 'px';
    }

    textarea.addEventListener('keydown', function (e) {
      // hide on enter
      // but don't hide on shift + enter
      if (e.keyCode === 13 && !e.shiftKey) {
        textNode.text(textarea.value);
        removeTextarea();
      }
      // on esc do not set value back to node
      if (e.keyCode === 27) {
        removeTextarea();
      }
    });

    textarea.addEventListener('keydown', function (e) {
      scale = textNode.getAbsoluteScale().x;
      setTextareaWidth(textNode.width() * scale);
      // textarea.style.height = 'auto';
      // textarea.style.height =
      //   textarea.scrollHeight + textNode.fontSize() + 'px';
    });

    function handleOutsideClick(e) {
      if (e.target !== textarea) {
        textNode.text(textarea.value);
        removeTextarea();
      }
    }
    setTimeout(function(){
      window.addEventListener('click', handleOutsideClick);
    });
  }.bind(textNode, tr, layer));

  layer.text_boxes.push(textNode);

  return textNode;
}

function destroy_text(target_node)
{
  target_node.children.forEach(function(child){
    child.destroy();
  })
  target_node.hidden_field_block.remove();
  target_node.destroy();
}