class GreetingCardCategorySetting
  include Mongoid::Document
  include Mongoid::Timestamps
  include ::GreetingCardSettingConcern

  DateTypes = ['regular_date', 'birthday']

  field :default_sort_number, type: Hash, default: {}
  
  field :automatically_send, type: Boolean, default: false
  field :auto_send_image_id, type: BSON::ObjectId
  field :auto_send_date_type, type: Integer, default: -1
  field :sender_mail, type: String
  field :regular_date, type: String
  field :recipient_ids, type: Array, default: []

  field :need_check_customs, type: Array, default: [] #From GreetingCardSetting
  field :reject_customs, type: Array, default: [] #From GreetingCardSetting
  field :agree_customs, type: Array, default: [] #From GreetingCardSetting

  field :category_id, type: String
  field :use_default, type: Boolean, default: false

  field :host_url
  field :update_user_id

  scope :enabled, ->{any_of([{:use_default.ne=>true}, {:use_default=>true, :default_sort_number.nin=>[nil, {}]}])}
  scope :custom_enabled, ->{where({:use_default.ne=>true})}

  before_save do
    unless @skip_callback
      if @custom_fields_changed
        @delete_custom_fields = @delete_custom_fields.select{|k| k.include?("default@")}
        self.delete_customs_func(@delete_custom_fields, true)
      end
      if self.automatically_send
        if !self.automatically_send_was || self.auto_send_image_id_changed? || self.auto_send_date_type_changed? || self.sender_mail_changed? || self.regular_date_changed?
          self.generate_records
        end
      end
      true # prevent return false in before_save
    end
  end
  def self.get_images(category_id)
    greeting_card_setting = GreetingCardSetting.first
    images = []
    if greeting_card_setting
      images += greeting_card_setting.greeting_card_images.map{|img| img.file.url}
    end
    if category_id
      greeting_card_setting2 = self.where(:category_id=> category_id).first
      if greeting_card_setting2
        images += greeting_card_setting2.greeting_card_images.map{|img| img.file.url}
      end
    end
    images
  end

  def agree_customs_func(apply_custom_fields, check_custom_fields=nil)
    if apply_custom_fields.count != 0
      self.agree_customs += apply_custom_fields
      self.agree_customs.uniq!
      if check_custom_fields.nil?
        check_custom_fields = GreetingCardSetting.first.custom_fields
      end
      apply_custom_fields.each{|k| self.custom_fields["default@#{k}"] = check_custom_fields[k]}
      self.need_check_customs -= apply_custom_fields
      self.reject_customs -= apply_custom_fields
      self.recalc_sort(false)
      @skip_callback = true
      self.save
      @skip_callback = false
    end
    true
  end
  def delete_customs_func(delete_custom_fields, add_to_reject=false)
    if delete_custom_fields.count != 0
      delete_custom_fields = delete_custom_fields.map{|f| f.sub(/^default@/, '')}
      self.need_check_customs -= delete_custom_fields
      self.agree_customs -= delete_custom_fields
      if add_to_reject
        self.reject_customs += delete_custom_fields
        self.reject_customs.uniq!
      else
        self.reject_customs -= delete_custom_fields
        delete_custom_fields.each do |f|
          self.custom_fields.delete("default@#{f}")
        end
      end
      self.recalc_sort(false)
      @skip_callback = true
      self.save
      @skip_callback = false
    end
    true
  end
  def update_need_check_customs(check_custom_fields, is_locale=false, need_check=false)
    if need_check
      check_custom_fields = check_custom_fields.map{|k,v| [k, v["field"][I18n.default_locale.to_s]]}.to_h unless is_locale
      category_custom_fields = self.custom_fields.map{|k,v| v["field"][I18n.default_locale.to_s]}
      self.need_check_customs += check_custom_fields.select{|k,v| v.blank? || !(category_custom_fields.include?(v))}.keys
    else
      self.need_check_customs += check_custom_fields.keys
    end
    self.need_check_customs = self.need_check_customs.uniq
    self.need_check_customs -= self.reject_customs
    self.need_check_customs -= self.agree_customs
    unless is_locale
      self.agree_customs.each{|k| self.custom_fields["default@#{k}"] = check_custom_fields[k]}
    end
    self.recalc_sort(false)
    @skip_callback = true
    self.save
    @skip_callback = false
  end

  def recipients
    MemberProfile.where(:id.in=> self.recipient_ids)
  end

  def enabled
    if use_default
      default_sort_number.present?
    else
      true
    end
  end

  def generate_record(site, module_app, member_profile, locale, sending_time, today)
    update_user = User.where(id: self.update_user_id).first
    if sending_time.nil?
      sending_time = member_profile.birthday
      sending_time = Date.parse(sending_time.strftime("#{today.year}-%m-%d"))
    end
    if sending_time < today
      sending_time += 1.year
    end
    # only update future email
    email_record = GreetingCardEmail.where(category_id: self.category_id, member_profile: member_profile, year: today.year, sending_time: {"$gt"=> today.to_datetime}, is_sent: false).first
    if email_record.nil?
      I18n.with_locale(locale) do
        email = Email.new
        email.save
        r = GreetingCardRecord.new(category_id: self.category_id, is_hidden: true, title: member_profile.name, name: member_profile.name, sender_mail: self.sender_mail, recipients: member_profile.email, sending_time: sending_time, card: self.auto_send_image_id)
        r.email_id = email.id
        r.save
        email_record = GreetingCardEmail.create(category_id: self.category_id, year: today.year, sending_time: sending_time, member_profile: member_profile, email: email, greeting_card_record: r)
        r.build_email(site, self.host_url, update_user, module_app, locale, email, {"model_class"=> "GreetingCardEmail", "model_id"=> email_record.id})
      end
    else
      r = email_record.greeting_card_record
      r.instance_variable_set(:@skip_callback, true)
      if r.card != self.auto_send_image_id # no need to generate image
        r.generate_image
      end
      r.update(sending_time: sending_time, card: self.auto_send_image_id)
      # I18n.with_locale(locale) do
      #   r.build_email(site, self.host_url, update_user, module_app, locale, r.email, {"model_class"=> "GreetingCardEmail", "model_id"=> email_record.id})
      # end
      email_record.update(sending_time: sending_time)
      r.email.update(mail_sentdate: sending_time)
    end
  end

  def generate_records
    layout_design = GreetingCardLayoutDesign.where(:category_id=> category_id).first
    if layout_design.nil?
      default_layout_design = GreetingCardLayoutDesign.where(:category_id=> nil).first
      if default_layout_design.nil?
        return false
      else
        layout_design = default_layout_design
      end
    end
    if @generating_records
      return false
    else
      @generating_records = true
      Thread.new do
        site = Site.first
        module_app = ModuleApp.where(key: 'greeting_card').first
        today = Date.today
        if self.auto_send_date_type == 0
          sending_time = DateTime.parse("#{today.year}/#{self.regular_date}")
        else
          sending_time = nil # use birthday
        end
        self.recipients.to_a.each do |member_profile|
          self.generate_record(site, module_app, member_profile, I18n.locale, sending_time, today)
        end
        @generating_records = false
      end
      return true
    end
  end

end
