module GreetingCard
  class Engine < ::Rails::Engine
    initializer "greeting_card" do
      Rails.application.config.to_prepare do
        locales = Site.first.in_use_locales rescue [:en,:zh_tw]
        begin
          t_data0 = locales.collect{|l| Dir["#{ENV['PWD']}/config/locales/#{l.to_s}.yml"]}.flatten.map{|yaml_file| YAML.load(File.read(yaml_file))}
          t_data1 = Dir["#{GreetingCard::Engine.root}/config/locales/*.yml"].map{|yaml_file| YAML.load(File.read(yaml_file))}
          data_item = {}
          key_item1 = {}
          value_item1 = {}
          value_item2 = {}
          t_data0.each do |t_data|
            v = t_data.values
            k = t_data.keys[0]
            key_item1[k] = v[0]['theme']
          end
          t_data1.each do |t_data|
            v = t_data.values
            k = t_data.keys[0]
            value_item1[k] = v[0]['greeting_card']['widget']['index']
            value_item2[k] = v[0]['greeting_card']['is_published']
          end
          data_item[key_item1] = [value_item1,value_item2]
        rescue  => e
          puts ['error in greeting_card',e]
        end
        begin
          require File.join(GreetingCard::Engine.root, 'app/models/greeting_card_setting')
          require File.join(GreetingCard::Engine.root, 'app/models/greeting_card_list_setting')
          require File.join(GreetingCard::Engine.root, 'app/models/greeting_card_category_setting')
          require File.join(GreetingCard::Engine.root, 'app/models/greeting_card_setting_index')
          if defined?(GreetingCardSetting)
            greeting_card_setting = GreetingCardSetting.first
            if greeting_card_setting.nil?
              greeting_card_setting = GreetingCardSetting.create()
            elsif GreetingCardSetting.count > 1
              GreetingCardSetting.all[1..-1].each do |a|
                a.destroy
              end
            end
            if greeting_card_setting.email_regex_enable
              GreetingCardCategorySetting.all.update_all(email_regex_enable: greeting_card_setting.email_regex_enable,email_regex: greeting_card_setting.email_regex)
            end
            GreetingCardCategorySetting.all.each do |greeting_card_category_setting|
              sort_number = greeting_card_category_setting.sort_number
              have_custom_key = false
              sort_number.keys.each do |sort_number_key|
                have_default_key ||= sort_number_key.include?("default@default@")
                have_custom_key ||= sort_number_key.include?("custom@")
              end
              if !have_custom_key
                new_sort_number = {}
                sort_number.keys.each do |sort_number_key|
                  new_sort_number_key = sort_number_key
                  if sort_number_key.include?("default@default@")
                    new_sort_number_key = sort_number_key.sub(/default@/, '')
                  else
                    new_sort_number_key = sort_number_key.sub(/default@/, "custom@#{greeting_card_category_setting.id}@")
                  end
                  new_sort_number[new_sort_number_key] = sort_number[sort_number_key]
                end
                greeting_card_category_setting.sort_number = new_sort_number
                greeting_card_category_setting.save
              end
            end
          end
          if defined?(GreetingCardListSetting)
            greeting_card_list_setting = GreetingCardListSetting.first
            if greeting_card_list_setting.nil?
              greeting_card_list_setting = GreetingCardListSetting.create()
            elsif GreetingCardListSetting.count > 1
              GreetingCardListSetting.where(:category_id=>nil)[1..-1].each do |a|
                a.destroy
              end
            end
          end
          if defined?(GreetingCardSettingIndex)
            if GreetingCardSettingIndex.first.nil?
              GreetingCardSettingIndex.create()
            elsif GreetingCardSettingIndex.count > 1
              GreetingCardSettingIndex.all[1..-1].each do |greeting_card_setting_index|
                greeting_card_setting_index.destroy
              end
            end
          end
        rescue => e
          puts ['GreetingCardSetting not found',e]
        end
        OrbitApp.registration "GreetingCard", :type => "ModuleApp" do
          module_label "module_name.greeting_card"
          base_url File.expand_path File.dirname(__FILE__)
          begin
            set_keyword_contstraints ['/greeting_cards']
          rescue => e
            puts ['the site is too old']
          end
          widget_methods ["widget"]
          widget_settings [{"data_count"=>10}]
          taggable "GreetingCardRecord"
          begin
            show_option_items data_item
          rescue => e
            puts ['there_was_no_show_option_method',e]
          end
          categorizable
          authorizable
          frontend_enabled
          # data_count 1..10

          side_bar do
            head_label_i18n 'module_name.greeting_card', icon_class: "icon-gift"
            available_for "users"
            active_for_controllers (['admin/greeting_cards'])
            head_link_path "admin_greeting_cards_path"

            context_link 'greeting_card.all_greeting_cards',
                                  :link_path=>"admin_greeting_cards_path" ,
                                  :priority=>1,
                                  :active_for_action=>{'admin/greeting_cards'=>'index'},
                                  :available_for => 'users'
            context_link 'greeting_card.setting',
                                  :link_path=>"setting_admin_greeting_cards_path" ,
                                  :priority=>2,
                                  :active_for_action=>{'admin/greeting_cards'=>'setting'},
                                  :available_for => 'managers'
            context_link 'greeting_card.category_setting',
                                  :link_path=>"category_setting_admin_greeting_cards_path" ,
                                  :priority=>3,
                                  :active_for_action=>{'admin/greeting_cards'=>'category_setting'},
                                  :available_for => 'sub_managers'
            context_link 'greeting_card.order_fields',
                                  :link_path=>"order_fields_admin_greeting_cards_path" ,
                                  :priority=>4,
                                  :active_for_action=>{'admin/greeting_cards'=>'order_fields'},
                                  :available_for => 'sub_managers'
            # context_link 'greeting_card.print_setting',
            #                       :link_path=>"print_setting_admin_greeting_cards_path" ,
            #                       :priority=>5,
            #                       :active_for_action=>{'admin/greeting_cards'=>'print_setting'},
            #                       :available_for => 'sub_managers'

            context_link 'greeting_card.layout_design',
                                  :link_path=>"layout_design_admin_greeting_cards_path" ,
                                  :priority=>5,
                                  :active_for_action=>{'admin/greeting_cards'=>'layout_design'},
                                  :available_for => 'sub_managers'

            context_link 'categories',
                                  :link_path=>"admin_module_app_categories_path" ,
                                  :link_arg=>"{:module_app_id=>ModuleApp.find_by(:key=>'greeting_card').id}",
                                  :priority=>7,
                                  :active_for_action=>{'admin/greeting_cards'=>'categories'},
                                  :active_for_category => 'GreetingCard',
                                  :available_for => 'managers'
            # context_link 'tags',
            #                       :link_path=>"admin_module_app_tags_path" ,
            #                       :link_arg=>"{:module_app_id=>ModuleApp.find_by(:key=>'greeting_card').id}",
            #                       :priority=>7,
            #                       :active_for_action=>{'admin/greeting_card'=>'tags'},
            #                       :active_for_tag => 'GreetingCard',
            #                       :available_for => 'managers'
            context_link 'greeting_card.acknowledgements',
                                  :link_path=>"admin_greeting_card_acknowledgements_path" ,
                                  :priority=>8,
                                  :active_for_action=>{'admin/greeting_cards'=>'acknowledgements'},
                                  :available_for => 'managers'
            # context_link 'greeting_card.admins',
            #                       :link_path=>"admin_greeting_card_admins_path" ,
            #                       :priority=>9,
            #                       :active_for_action=>{'admin/greeting_cards'=>'admins'},
            #                       :available_for => 'managers'
            # context_link 'greeting_card.exports',
            #                       :link_path=>"export_admin_greeting_cards_path" ,
            #                       :priority=>10,
            #                       :active_for_action=>{'admin/greeting_cards'=>'exports'},
            #                       :available_for => 'managers'
          end

        end
      end
    end
  end
end